/********************************************************************
 *
 *  Command Line interpreter for Steuerung 13.06.2013
 *
 *  R. Kohleick
 *
 *  licensed under GNU GPLv3
 *
 ********************************************************************/


#include "Arduino.h"
#include <avr/pgmspace.h>
#include <Menus.h>
#include <mystuff.h>
#include <streeng.h>
#include <Controler.h>
#include <CommandLine.h>

#define EEPROMOFFSET 300

#define CLI_STATUS      0
#define CLI_TEMPALARM	1
#define CLI_CYCLE		2
#define CLI_ECONOMY     3
#define CLI_COMFORT     4
#define CLI_TEMPERATURE	5
#define CLI_ANTIFREEZE	6
#define CLI_PROFILE     7
#define CLI_ON          8
#define CLI_OFF         9
#define CLI_HELP		10
#define CLI_SET         11
#define CLI_RESET		12
#define CLI_TIME		13
#define CLI_COOLER      14
#define	CLI_DATE        15
#define CLI_LOG         16

#define CLI_STRINGS     17

#define EEPSTART	19	// CLI starts with entry 19 in string storage


/*****************************
 *
 * initializes commandline interpreter instance
 *
 *****/

CLI::CLI(void) {
}


/*****************************
 *
 * reads line terminated by CR into buffer
 * eliminates CR / LF pairs
 *
 * returns length of read string
 *
 *****/


int CLI::Readln(char *b) {
	byte i=0;
	byte c=0;
	
  	do {
		if(Serial.available()) {
			*b++=c=Serial.read();
			i++;
		}
	} while ((c!=';') && (c!='\r') && (i<140));
	
	*b=(char)0;
}


/****************************************************
 *
 * splits command line to atoms
 *
 * returns number of args identified and modifies pointers to atoms
 * pointers point to zero terminated strings
 *
 ******/

byte CLI::parseatoms(char *line, char *at[]) {
	byte number;
	char c, *p;
	
	number=0;
	p=line;
    
	do {
		while(isspace(*p) || ispunct(*p))
			p++;
		at[number]=p;
		while (isalnum(*p) || (*p=='.') || (*p==':')) {
			c=toupper(*p);
			*p++=c;
		}
		
		if(*p)
			*p++=0;
		
		if(*at[number])
			number++;
	} while ((*p) && (number<menumaxarg));
	
	return number;
}


/***************************************************
 *
 * checks whether string contains only digits
 *
 * returns true/false
 *
 ******/


boolean CLI::truenumber(char *s) {
	char *p;
	
	for(p=s; isdigit(*p); p++);
	
	return !(*p);
}

/***************************************************
 *
 * checks whether string contains only digits and
 * one occurence of special character
 *
 * returns true/false
 *
 ******/


boolean CLI::onlydigits(char *s, char special) {
	char *p;
	byte found=0;
	
	for(p=s; isdigit(*p) || (*p==special); p++)
		if (*p==special)
			found++;
	return (!(*p) && (found==1));
}

/***************************************************
 *
 * normalizes number with digit (temperature) to
 * one decimal accuracy if decimal.
 * expects that special char '.' is in string
 *
 ******/


void CLI::normalize(char *s) {
	char *p;
	byte found=0;
	
	for(p=s; *p !='.'; p++);
    p++;
    p++;
    *p=0;
}


/****************************************************
 *
 * Read the parameters into the parameter array
 *
 ******/

unsigned long CLI::readparams(char *args[], byte number, int *params) {
	
	byte i, j, k;
	char *p;
	byte thistype;
	int t;
	unsigned long type=0;
	char buffer[20];
    
    streeng eeprom(EEPROMOFFSET);
    eeprom.setbaseindex(EEPSTART);
	
	for(k=0; k<number; k++) {
		i=number-k-1;
		
		type<<=4;
		thistype=NOTVALID;
		
		for (j=0; j<CLI_STRINGS && thistype==NOTVALID; j++) {
			
			eeprom.read(buffer,j);
            
			if (strcmp(args[i],buffer)==0) {
				params[i]=j;
				thistype=KEYWORD;
			}
			
			if(params[i]==CLI_ON) {
				thistype=ONOFF;
				params[i]=1;
			}
			
			if(params[i]==CLI_OFF) {
				thistype=ONOFF;
				params[i]=0;
			}
			
			if(params[i]==CLI_SET) {
				thistype=SET;
				params[i]=1;
			}
			
			if(params[i]==CLI_RESET) {
				thistype=RESET;
				params[i]=1;
			}
            
			if(params[i]==CLI_LOG) {
				thistype=LOG;
				params[i]=1;
			}
		}
		
            // check for time
        if (thistype==NOTVALID && onlydigits(args[i],':')) {
            thistype=TIME;
            eliminate(args[i],':');
            t=atoi(args[i]);
            params[i]=t%100;
            t/=100;
            params[i]+=t*60;
            if (params[i]>1440) {   // error, wrong time format
                params[i]=0;
                thistype=NOTVALID;
                goto error;
            }
        }
		
            // check for temp value
        if (thistype==NOTVALID && onlydigits(args[i],'.')) {
            thistype=VALUE;
            normalize(args[i]);
            eliminate(args[i],'.');
            params[i]=atoi(args[i]);
		}
		
            // check for integer
        if (thistype==NOTVALID && truenumber(args[i])) {
            thistype=INTEGER;
            params[i]=atoi(args[i]);
		}
		
    error:
        type|=thistype;
	}
    
	if ((type & 0x0F)!=KEYWORD)
		type=ERROR;
	else
		type = (type & 0xFFFFFFF0) | ( params[0] & 0x0F);
	
	return type;
}

/****************************************************
 *
 * Check Syntax
 *
 * checks for correct syntax and converts the syntax to menu entry numbers
 *
 ******/


prog_uint32_t intcode[] PROGMEM = {
    0x00000000,	// status							8000
    0x00000022,	// cycle on/off						8001
    0x00434352,	// cycle set time temp time temp	8002
    0x00000023,	// economy on/off					8003
    0x00000453,	// economy set temp					8004
    0x00000024,	// comfort on/off					8005
    0x00000454,	// comfort set temp					8006
    0x00000025,	// temperature on					8007
    0x00000455,	// temperature set temp				8008
    0x00000026, // antifreeze on					8009
    0x00000456,	// antifreeze set temp				800A
    0x00000667,	// profile nr integer (0=Never, 12=Always)	800B
    0x00043567,	// profile nr set time temp			800c
    0x00000767,	// profile nr reset					800d
    0x0000000A,	// help no args						800e
    0x0000035d,	// time set time					800f
    0x0000002e,	// cooler on / off					8010
    0x00000261,	// alert nr on/off					8011
    0x00004561,	// alert nr set temp				8012
    0x0006665F, // date set dd mm yy				8013
    0x00000001, // alert							8014
    0x00003580,	// status log set time				8015
    0x00000280	// status log on / off				8016
};


const byte numberofcommands=23;

unsigned int CLI::SyntaxCheck(unsigned long type, int *args) {
    
	unsigned int command=SYNTAXERROR;
	byte i;
    
	for(i=0;i<numberofcommands && command==SYNTAXERROR; i++) {
        
		if(pgm_read_dword( & intcode[i] )==type) {
			command=0x8000+i;
		}
	}
    
	switch(command) {
		case 0x8002:		//	cycle set time temp time temp
			if(args[2]>=args[4])
				command=SYNTAXERROR;
			break;
            
		case 0x8011:		// tempalarms 0, 1
		case 0x8012:
			if(args[1]>1)
				command=SYNTAXERROR;
			break;
            
		default:
			break;
    }
	return command;
}



/****************************************************
 *
 * Interpret Command Line
 *
 * returns command as in menu tree commands
 *
 ******/

unsigned int CLI::interpret(char *line, int *args) {
	char *atom[menumaxarg];     // menumaxarg is supplied by menu package
	unsigned long type;
	unsigned int command;
	byte argc;
	
	argc=parseatoms(line, atom);
    
	type=readparams(atom, argc, args);
    
	command=SyntaxCheck(type, args);
	
	*line=0;	// force delete of line
	return command;
}
