/********************************************************************
 *
 *  Controler - controls heater and does switching, timing etc.
 *  for Steuerung 13.06.2013
 *
 *  R. Kohleick
 *
 *  licensed under GNU GPLv3
 *
 ********************************************************************/



#include "Arduino.h"

#include <Menus.h>
#include <myLiquidCrystal.h>
#include <avr/pgmspace.h>
#include <EEPROM.h>
#include <mystuff.h>
#include <streeng.h>
#include <Controler.h>
#include <RTC_DS2415.h>

extern myLiquidCrystal lcd;
extern Clock rtclock;

int const eepromoffset=300;

#define EEPROMINDEX	0


/***********************
 *
 * checks whether there is a risk of malfunction from the heater device
 *
 ******/

boolean Control::checkformalfunction(int max) {

	if(!heaterstatus)                                   // heater is not running, so we cannot check whether it works
		return(false);

	if(rtclock.now()<laston+maxwaitfortempincrease)		// minimum runtime for heater required, we cannot say whether malfunction,
		return(false);                                  // if heater did not run long enough for the test
                                                        // from here on we know that heater runs for more than maxwaitfortempincrease seconds
	return max<starttemp+2*hysteresis;                  // since last heaterstart, temperature did not move up, i.e. less than 2 x hysteresis
}


/*************
 *
 * initializes min max temperature data in Controler
 *
 ***/

void Control::initminmax(int inside, int outside){

    in_tempmax=in_tempmin=inside;
    out_tempmin=out_tempmax=outside;
}


/*************
 *
 * updates temperature data to Controler
 *
 ***/

void Control::update(int inside, int outside, byte sns){

	in_temp=inside;
	out_temp=outside;
	sensors=sns;
    in_tempmin=min(in_temp,in_tempmin);
    in_tempmax=max(in_temp,in_tempmax);
    out_tempmin=min(out_temp,out_tempmin);
    out_tempmax=max(out_temp,out_tempmax);
}

/*************
 *
 * determines, what the current temp setpoint should be
 *
 **/

int Control::temp_soll(int time, byte dow){
	byte i;
	int solltemp;

	switch(mode) {

		case CTRL_MANUAL:
			solltemp=manualtemp;
			break;

		case CTRL_CMFT:
			solltemp=comfort;
			break;

		case CTRL_ECON:
			solltemp=economy;
			break;

		case CTRL_CYCLE:
			solltemp=economy;
			if((time>=comfort_on) && (time<comfort_off))
				solltemp=comfort;
			break;

		case CTRL_SCHEDULE:                        // find first schedule active at this point in time and return temp
			for(i=0;i<MAXPROFILE;i++){

				if(profile[i].isactivetoday(dow)) {
					solltemp= profile[i].findsoll(time);
					i=MAXPROFILE;
				}

			}
			break;

		case CTRL_OFF:
		case CTRL_AFREEZE:
		default:
			solltemp=antifreeze;
			break;
    }
	return max(solltemp,antifreeze);	// make sure always > antifreeze
}


/**************************************
 *
 * determines, if schedule is active
 *
 *****/

boolean Schedule::isactivetoday(byte dow) {

	switch(active) {

		case ALWAYS:
			return(true);
			break;


		case WORKDAYS:
			return ((dow>SUNDAYS) && (dow<SATURDAYS));
			break;

		case WEEKENDS:
			return ((dow==SUNDAYS) || (dow==SATURDAYS));
			break;

		case FIRSTHALF:
			return ((dow>SUNDAYS) && (dow<THURSDAYS));
			break;

		case SECONDHALF:
			return ((dow>WEDNESDAYS) && (dow<SATURDAYS));
			break;

		case SUNDAYS:
		case MONDAYS:
		case TUESDAYS:
		case WEDNESDAYS:
		case THURSDAYS:
		case FRIDAYS:
		case SATURDAYS:
			return (dow==(active-SUNDAYS));
			break;

		case NEVER:
		default:
			return(false);
			break;
	}
}


/*****************************************************
 *
 * determines for a given schedule, the temperature indicated by the schedule
 *
 *
 *******/

int Schedule::findsoll(int time) {
	byte i;

	int soll=setpoints[0].temp;

	for(i=1; i<number; i++) {
		if (time>=setpoints[i].time)
			soll=setpoints[i].temp;
	}

	return soll;
}


/***************************************************
 *
 *
 * allows dynamic generation of menus
 *
 * function passed to and called from the Menu::display routine
 * copies string, which can be created dynamically, to menu item.
 * Instead of a static string residing in Progmem
 *
 *****/

Schedule *_sch;
Menu *_menu;

void _cpydynamicmenu(byte ident, char *s) {
	int pos;

	streeng eep(eepromoffset);
	eep.setbaseindex(EEPROMINDEX);

	if(ident==1) {	// first one is activity
		eep.read(s,_sch->active);

		return;
	}

	pos=ident-2;	// make pos equal to index in schedule.setpoints[]
					// following are actual schedule entries

	sprinttime(s, mod2hrs(_sch->setpoints[pos].time), mod2min(_sch->setpoints[pos].time));
	s[5]=' ';
    sprinttemp(&s[6],_sch->setpoints[pos].temp);
}


/*****************************************************
 *
 * create a menu from the schedule
 *
 * m - pointer to menu
 *
 *******/

void Schedule::menufromschedule(Menu *m) {

	int i;
	char s[MENUSTRLEN], t[MENUSTRLEN];
	unsigned int level;

	_sch=this;
	_menu=m;

	m->menureset();

	m->entry(0x1000, 0x81);    // 1 st dynamic = current activity of schedule

	m->entry(0x1100, NEVER);
	m->entry(0x1200, SUNDAYS);
	m->entry(0x1300, MONDAYS);
	m->entry(0x1400, TUESDAYS);
	m->entry(0x1500, WEDNESDAYS);
	m->entry(0x1600, THURSDAYS);
	m->entry(0x1700, FRIDAYS);
	m->entry(0x1800, SATURDAYS);
	m->entry(0x1900, WORKDAYS);
	m->entry(0x1A00, WEEKENDS);
	m->entry(0x1B00, FIRSTHALF);
	m->entry(0x1C00, SECONDHALF);
	m->entry(0x1D00, ALWAYS);

	for(i=0; i<number; i++) {

		level=0x2000+i*0x1000;
		m->entry(level, 0x80+i+2);	// make this dynamic
		lastentrytoedit=level;		// remember last level, in case event was added, this is new event

		m->entry(level+0x0100, 13);    // edit temp
		if(i>0) {                      // first entry always remains, will always start at 00:00
			m->entry(level+0x0200, 14);    // edit time
			m->entry(level+0x0300, 15);    // delete
		}
	}

	level=0x2000+number*0x1000;
	m->entry(level, 16); // "Add Event"

	level=0x2000+(number+1)*0x1000;
	m->entry(level, 17); // "return to main Menu"

	m->obtainstring(t,18);

	m->setpanel(t);
	m->home();
}


/*****************************************************
 *
 * allows editing of schedule
 *
 * m - pointer to menu
 *
 *******/

boolean Schedule::editschedule(void) {

	int i;
	byte hrs, min;
	unsigned int command;
	boolean quit=false;
	byte initialize=2;
    const byte menusize=13+4*MAXSETPOINTS+2+1;       // 13 time strings, 1+3 items per setpoint, 2 (add+main menu)
	menuitem ctrlmenu[menusize];

	streeng	eep(eepromoffset);
	eep.setbaseindex(EEPROMINDEX);

        // create a menu instance
	Menu schedulemenu(ctrlmenu, menusize, &eep, 0, _cpydynamicmenu);

        // initialize buttons, sequence: up, enter, down
	schedulemenu.begin(11, 10, 9);

	do {
		menufromschedule(&schedulemenu);

		if(initialize<2) {
			switch(initialize) {
				case 0:
					command=lastentrytoedit+0x100;
					initialize=2;
					break;
				case 1:
					command=lastentrytoedit+0x200;
					initialize--;
					break;
			}
		}
		else
			command=schedulemenu.event();

		if((command>=0x1100) && (command<=0x1D00)) {
			active=(command-0x1100)>>8;
		}

		if((command>=0x2000) && (command<0x2000+0x1000*number)) {   // we ave received notice that an event needs to be changed

			i=(command>>12) - 2;                     // determine which entry to work on

			switch(command & 0x0FFF) {                  // now look for submenus
				case 0x100:  // edit the entry temperature
					setpoints[i].temp=schedulemenu.getint(setpoints[i].temp, 0, 500, 4, 1, "Temp ? ", 1, 2);
					break;

				case 0x200:  // edit the entry time
					hrs=mod2hrs(setpoints[i].time);
					min=mod2min(setpoints[i].time);
					schedulemenu.gettime(&hrs, &min, "Time ? ", 1, 2);
					setpoints[i].time=mod(hrs,min);
					sort();
					break;

				case 0x300:  // delete the entry
					del(i+1); // we do not delete array index, but number
					break;
			}
		}

		if(command==0x2000+0x1000*number) {
			add(setpoints[number-1].temp, mod2hrs(setpoints[number-1].time), mod2min(setpoints[number-1].time));
			initialize=1;
		}

		if(command==0x2000+0x1000*(number+1))
			quit=true;

	} while (!quit);

	return (active!=NEVER);
}


/*****************************************************
 *
 * Constructor
 *
 *******/

Schedule::Schedule(int temp) {

	number=0;
	add(temp, 0, 0);  //  all scehdules start with 00:00 hrs
}


/*****************************************************
 *
 * add a timing event
 *
 * t - the temperature
 * h - the hours, when the temperature is to be valid
 * m - the minutes ..
 *
 *******/

void Schedule::add(int t, byte h, byte m) {

	if(number<MAXSETPOINTS) {
		setpoints[number].temp=t;
		setpoints[number].time=mod(h,m);
		number++;
		sort();
	}
}


/*****************************************************
 *
 * delete a timing event
 *
 * n - the event number, events are numbered 1...number, whereas stored 0... number-1
 *
 *******/

void Schedule::del(byte n) {

	byte i;

	if(number<2)
		return;

	for(i=n-1; i<number-1; i++) {
		setpoints[i].temp=setpoints[i+1].temp;
		setpoints[i].time=setpoints[i+1].time;
	}
	number--;
	sort();
}


/*****************************************************
 *
 * sorts timing events by time (ascending)
 *
 *
 *******/

void Schedule::swap(int *i, int *j) {
	int t;

	t=*i;
	*i=*j;
	*j=t;
}




void Schedule::sort() {

	byte i, j;

	for(i=0; i<number; i++) {
		for(j=i+1; j<number; j++) {
			if(setpoints[i].time > setpoints[j].time) {
				swap(&setpoints[i].time, &setpoints[j].time);
				swap(&setpoints[i].temp, &setpoints[j].temp);
			}
		}
	}
}


/***********************************
 *
 *  creates a Control instance
 *
 */

Control::Control(void) {

	Controlreset();
}


/***********************************
 *
 *  intializes outputpins
 *
 */

void Control::begin(int heater) {

    heaterpin=heater;

	pinMode(heaterpin,OUTPUT);
	heaterstatus=LOW;

	digitalWrite(heaterpin,heaterstatus);
}


/***********************************
 *
 *  creates a Control instance
 *
 */

void Control::Controlreset(void) {

	heaterstatus=LOW;

	mode=CTRL_OFF;

	hysteresis=1;
	heaterminruntime=30;

	logtime=-mod(9,00); // default logging is at 09h00, neg=switch it off
	logging=false;
    malfunction=false;

	laston=0;

	comfort=200;		// make comfort temp level 20.0 deg C
	economy=170;		// make economy temp level 17.0 deg C
	antifreeze=80;		// fix freezeprotection to 8.0 deg C
	manualtemp=comfort;

    in_tempmax=out_tempmax=-1000;
    in_tempmin=out_tempmin=1000;

	comfort_on=mod(7,00);	// switch comfort on at 7h00
	comfort_off=mod(22,00);	// switch to economy at 22h00

    deleteallschedules();
}



/***********************************
 *
 *  delete all schedules in controler
 *
 ****/

void Control::deleteallschedules(void) {
    byte i;

    for(i=0;i<MAXPROFILE;i++) {
        profile[i].number=0;
        profile[i].add(antifreeze,0,0);
        profile[i].active=NEVER;
    }
}


/***********************************
 *
 *  disable all schedules in controler
 *
 ****/

void Control::disableallschedules(void) {
    byte i;

    for(i=0;i<MAXPROFILE;i++)
        profile[i].active=NEVER;
}


/***********************************
 *
 *  determine, whether any scheduel is active at all
 *
 *****/

boolean Control::isanyscheduleactive(void) {
    byte i;

    for(i=0;i<MAXPROFILE;i++) {
        if(profile[i].active!=NEVER)
            return true;
    }
    return false;
}


/*************************************************
 *
 * regulate heater
 *
 * performs  heater regulation with
 * hysteresis and
 * heater minimum runtime
 *
 * returns:
 *		soll - target temperature
 *
 ***************/

int Control::regulate() {

    unsigned long now;
    static int lasttime;        //  remembers last time that this function was called (in sec from rtc)
    static int maxtemp;         //  maximum temperature since last heater start
	int time;                   //  time in minutes of day, used for determining time in schedules

	boolean startheater, stopheater;

	int soll;

    now=rtclock.now();                              // now holds absolute time in seconds
    time=mod(rtclock.hours(), rtclock.minutes());   // mod = minutes of day, format by which schedules are organized

    logging=(logtime==time && time!=lasttime);
    lasttime=time;

    soll=temp_soll(time, rtclock.dow());

	if(!heaterstatus) {
		startheater= (in_temp<(soll-hysteresis));	// switch on heater
		if(startheater) {
			heaterstatus=HIGH;                      // possibly override coolerminruntime if temp too cool.
			laston=now;                             // remember when switched on.
			starttemp=maxtemp=in_temp;              // and what was the starting temperature, which is also the init for the max temperature since last heaterstart
		}
	} else {
		stopheater=(in_temp>(soll+hysteresis));
		stopheater=stopheater && ((now-laston)>(long)heaterminruntime);
		if(stopheater) {
			heaterstatus=LOW;
		}
	}

    digitalWrite(heaterpin,heaterstatus);           // write new heaterstatus to the pin, does not matter, if nothing is changed

    if (in_temp>maxtemp)
        maxtemp=in_temp;
    
	malfunction=checkformalfunction(maxtemp);
	return soll;
}


/***********************************
 *
 *	print the controler Status into string s
 *
 *  s - buffer of minimum lenght 140
 *  len - maximum stringlength
 *
 */

void Control::status(char *s, byte len) {
	char t[12];
    byte i;

    rtclock.updateclock();

	strcpy(s,"Date: ");
    rtclock.sprintdate(t);
    mystrcat(s,t,len);
    mystrcat(s," ",len);
    rtclock.sprinttime(t);
    mystrcat(s,t,len);
    mystrcat(s," ",len);

	mystrcat(s,"\nIn T: ",len);
    sprinttemp(t, in_temp);
    mystrcat(s,t,len);

    mystrcat(s," (",len);
    sprinttemp(t, in_tempmin);
	mystrcat(s,t,len);
	mystrcat(s," ",len);
    sprinttemp(t, in_tempmax);
	mystrcat(s,t,len);


	mystrcat(s,")\nSetP : ",len);
    sprinttemp(t,temp_soll(mod(rtclock.hours(),rtclock.minutes()), rtclock.dow()));
	mystrcat(s,t,len);

	if(sensors>1) {
		mystrcat(s,"\nOutT: ",len);
        sprinttemp(t, out_temp);
		mystrcat(s,t,len);
        mystrcat(s," (",len);
        sprinttemp(t, out_tempmin);
        mystrcat(s,t,len);
        mystrcat(s," ",len);
        sprinttemp(t, out_tempmax);
        mystrcat(s,t,len);
        mystrcat(s,")",len);
	}

	strcat(s,"\nMode: ");
    switch(mode) {
        case 'M':
            strcpy(t,"Man");
            break;

        case 'C':
            strcpy(t,"Cmf");
            break;
            
        case 'E':
            strcpy(t,"Eco");
            break;
            
        case 'O':
            strcpy(t,"Cyc");
            break;
            
        case ' ':
            strcpy(t,"OFF");
            break;
            
        case 'A':
            strcpy(t,"Afz");
            break;
            
        case 'P':
            strcpy(t,"Prg");
            break;
    }
	mystrcat(s,t,len);
    
	mystrcat(s,"\nHeat: ",len);
	if (heaterstatus)
		mystrcat(s,"ON",len);
	else
		mystrcat(s,"OFF",len);
    
	mystrcat(s,"\nLog: ",len);
	if (logtime<0)
		mystrcat(s,"OFF",len);
	else {
		i=logtime/60;
		t[0]='0'+i/10;
		t[1]='0'+i%10;
		t[2]=':';
		i=logtime%60;
		t[3]='0'+i/10;
		t[4]='0'+i%10;
		t[5]=0;
		mystrcat(s,t,len);
	}
}
