/***************************
 *
 *
 *	Library to create an interactive LCD screen menu system
 *
 *
 *	R. Kohleick	29.12.2011
 *
 * licensed under GNU GPLv3
 *
 **/


#include <myLiquidCrystal.h>
#include <Menus.h>
#include <mystuff.h>
#include <streeng.h>
#include <avr/pgmspace.h>


extern myLiquidCrystal lcd;

#define STARTPOS	0xff


/**********************************************************************
 *
 * Prepares the editing of string
 *
 *	x - column position
 *	y - line position
 *	q - question string
 *	i - item to be edited
 *
 **********/

void Menu::prepare_edit(byte *x, byte *y, char *q, char *i) {
	displaylocked=true;

	cleararound(*y);
	forceredraw();

	lcd.setCursor(*x,*y);
	lcd.print(q);
	*x+=strlen(q);

	lcd.print(i);
	lcd.blink();
}


/**********************************************************************
 *
 * normalizes chars in timestring
 *
 ********/

void Menu::normaltimechar(char c, char *s, byte pos) {

    switch(pos) {
        case 0:
            c=inrange(c,'0','2');
            break;

        case 1:
            c=inrange(c,'0','9');
            if(s[0]=='2')
                c=min(c,'3');
            break;

        case 3:
            c=inrange(c,'0','5');
            break;

        case 4:
            c=inrange(c,'0','9');
            break;
	}
    s[pos]=c;
}


/**********************************************************************
 *
 * gets a time value from screen
 *
 * hours - ptr to hours
 * minutes - ptr to minutes
 * question - the question to ask
 * x - column position of string
 * y - line position of strung
 *
 *	return true if time changed, false if time remains unchanged
 *
 **********/


boolean Menu::gettime(byte *hours, byte *minutes, char* question, byte x, byte y) {

	char s[6];
	int time;
	byte c, pos;
	boolean quit=false;
	boolean redrawchar=false;
	boolean nochange=false;
	const char tms=':';

	sprinttime(s,*hours, *minutes);
	prepare_edit(&x, &y, question, s);

	pos=0;

	do {

		if(redrawchar) {
			lcd.setCursor(x+pos,y);
			lcd.write(s[pos]);
			redrawchar=false;
		}

		lcd.setCursor(x+pos,y);
		switch(getkey()) {
			case UPKEY:
				c=s[pos]+1;
				normaltimechar(c,s,pos);
				redrawchar=true;
				break;

			case DOWNKEY:
				c=s[pos]-1;
				normaltimechar(c,s,pos);
				redrawchar=true;
				break;

			case LONGUPKEY:
			case XTRALONGUPKEY:
				pos=skipright(s, pos, tms);
				break;

			case LONGDOWNKEY:
			case XTRALONGDOWNKEY:
				pos=skipleft(s, pos, tms);
				break;

			case ENTERKEY:
				quit=true;
				break;

			case LONGENTERKEY:
			case XTRALONGENTERKEY:
				nochange=true;
				quit=true;
				break;
		}

	} while (!quit);
	lcd.noBlink();

	displaylocked=false;

	if(!nochange) {
		eliminate(s,tms);
		time=atoi(s);

		*hours=time/100;
		*minutes=time%100;

		*hours=min(*hours,23);
		*minutes=min(*minutes,59);
	}

	return !nochange;
}


/**********************************************************************
 *
 * gets a an element of enumerated class from screen
 *
 * initial - the initial value
 * low - the low value
 * high - the max value
 * question - the question to ask
 * ss - array of charptrs pointing to names of enum, user has to ensure that all strings have equal length
 * x - column position of string
 * y - line position of strung
 *
 **********/

int Menu::getenum(int initial, int low, int high, char *ss[], byte x, byte y) {

	byte selection;
	boolean quit=false;
	boolean redrawchar=true;
	char s[MENUSTRLEN];

	initial=max(0,initial);
	initial=min(high,initial);

	displaylocked=true;

	lcd.setCursor(x,y);
	lcd.blink();

	selection=initial;
	do {

		if(redrawchar) {
			lcd.setCursor(x,y);

			strcpy(s,ss[selection]);
			lcd.print(s);
			redrawchar=false;
		}

		lcd.setCursor(x,y);
		switch(getkey()) {
			case UPKEY:
				selection=inrange(selection+1,low,high);
				redrawchar=true;
				break;

			case DOWNKEY:
				selection=inrange(selection-1,low,high);
				redrawchar=true;
				break;

			case ENTERKEY:
				quit=true;
				break;

			case LONGENTERKEY:
			case XTRALONGENTERKEY:

				selection=initial;
				quit=true;
				break;
		}

	} while (!quit);
	lcd.noBlink();

	displaylocked=false;
	return selection;
}


/**********************************************************************
 *
 *
 * private function, moves cursor to right within number string including dec
 *
 *
 **********/

byte Menu::skipright(char *s, byte pos, char c) {
	byte i;

	i=pos+1;
	while ((s[i]==c) && (s[i]))
		i++;

	if (!(s[i]))
		return(pos);

	return (i);
}

/**********************************************************************
 *
 * service function to be executed while waiting for keyboard entries,
 * calls the function provided by the external programs
 *
 * stores result in servicerequest variable, for future processing
 *
 **********/

void Menu::externalservice(void) {

    if (servicerequest!=NOCOMMAND)          // last servicerequest should not be overridden
        return;

    testtoswitchoffbacklight();             // check, whether we should turn off the backlight, executed frequently

    if (service!=0)
        servicerequest=(*service)(args);    // else take the command passed by the externalservice routine
}

/**********************************************************************
 *
 * senses for pin pressed
 *
 * pin - the pin it is sensing
 * key - the key that is pressed
 *
 * returns the key - or - if key is pressed for a longer period key + LONGKEYOFFSET
 *
 **********/

byte Menu::sensepin(byte pin, byte key) {

	unsigned int duration;

	delay(50);
	duration=millis();

	while(!digitalRead(pin))	// i.e. while key is pressed
		externalservice();      // regularly perform the menu service in idle times

    switchonbacklight(30);      // keypressed, switch on backlight
    duration=millis()-duration; // measure keypress

    delay(200);                 // debounce delay
	if (duration<500)
		return key;                 // normal = short keypress
	if (duration<1500)
		return key+LONGKEYOFFSET;	// long keypress
	return key+XTRALONGKEYOFFSET;	// extra long keypress
}


/**********************************************************************
 *
 *  waits for a key if pressed and returns the key
 *
 **********/

byte Menu::getkey() {
	byte key=NOKEY;

	do {
		externalservice();	// regularly perform the menu service in idle times

        if(digitalRead(uppin)==LOW)
            return sensepin(uppin,UPKEY);

        if(digitalRead(downpin)==LOW)
            return sensepin(downpin,DOWNKEY);

        if(digitalRead(enterpin)==LOW)
            return sensepin(enterpin,ENTERKEY);

    } while((key==NOKEY) && (servicerequest==NOCOMMAND));		// wait for a key pressed and released
																// or external service request issued
	return key;
}



/**********************************************************************
 *
 * displays and allows editing of integer number
 *   low - minimum value, will reject input if lower
 *   high - maximum value, will reject input if higher
 *   question - string to be printed before the integer field
 *   field - minimum field width to display
 *   dec - position of fake decimal
 *   x - column position at which to display number
 *   y - line position at which to display number (leftmost char)
 *
 **********/


int Menu::getint(int initial, int low, int high, byte field, byte dec, char *question, byte x, byte y)
{
	char s[MENUSTRLEN];
	unsigned int result;

	s[0]=0;

	prepare_edit(&x, &y, question, s);

	do {
		sprintint(initial, s, field, dec);
		leadingzeros(s);
		editnumstr(s, x, y,'.');
		eliminate(s,'.');
		result=atoi(s);
	} while ((result<low) || (result>high));

	return result;
}


/**********************************************************************
 *
 *
 * forces the int c to the range of ints between min or max
 * with a 'rollover' else, returns c
 *
 **********/

int Menu::inrange( int c, int cmin, int cmax) {

	if(c>cmax)
		return cmin;

	if(c<cmin)
		return cmax;

	return(c);
}

/**********************************************************************
 *
 *  prompts user to enter the month
 *
 * start - initial value
 * x - columnposition
 * y - lineposition
 *
 *********

 byte Menu::getmonth(byte start, byte x, byte y) {

 start=min(start,11);
 start=max(start,0);
 start=getenum(start, 0, 11, _monthofyear, x, y);

 return start;
 }
 */


/**********************************************************************
 *
 * evaluates date and determines, if correct
 *	day - day
 *	mon - month
 *	year - jahr
 *
 **********/

boolean Menu::correctdate(byte day, byte mon, int year) {

	byte dmax;

	switch(mon) {
		case 1:
		case 3:
		case 5:
		case 7:
		case 8:
		case 10:
		case 12:
			dmax=31;
			break;

		case 2:
			dmax=(year%4==0)?29:28;
			break;

		default:
			dmax=30;
			break;

	}

	return dmax>=day;
}


/**********************************************************************
 *
 * displays and allows editing of date string
 *	day - ptr to integer day
 *	mon - ptr to integer month
 *	year - ptr to integer year
 *	question - string to be printed before the integer field
 *	x - column position at which to display number
 *	y - line position at which to display number (leftmost char)
 *
 *	return true if date changed, false if date remains unchanged
 *
 **********/



boolean Menu::getdate(byte *day, byte *mon, int *year, char *question, byte x, byte y)
{
	char s[MENUSTRLEN];
	byte i, mm, dd;
	const char dsp='.';
	boolean changed;

	int yy, y1;

	s[0]=0;

	prepare_edit(&x, &y, question, s);

	do {

		dd=*day;
		mm=*mon;
		yy=y1=*year;

		s[0]='0'+dd/10;
		s[1]='0'+dd%10;
		s[2]=dsp;
		s[3]='0'+mm/10;
		s[4]='0'+mm%10;
		s[5]=dsp;
		s[6]='0'+y1/1000;
		y1%=1000;
		s[7]='0'+y1/100;
		y1%=100;
		s[8]='0'+y1/10;
		s[9]='0'+y1%10;
		s[10]=0;

		editnumstr(s, x, y, dsp);

		s[2]=0;
		s[5]=0;

		dd=min(atoi(s),31);
		mm=min(atoi(&s[3]),12);
		yy=min(atoi(&s[6]),2100);
		yy=max(yy,2000);

	} while(!correctdate(dd,mm,yy));

	displaylocked=false;

	changed=((*day!=dd) || (*mon!=mm) || (*year!=yy));

	*day=dd;
	*mon=mm;
	*year=yy;

	return changed;
}



/**********************************************************************
 *
 *
 * private function, moves cursor to left within number string including dec
 *
 *
 **********/

byte Menu::skipleft(char *s, byte pos, char c) {
	int i;

	if(pos==0)
		return 0;

	i=pos-1;

	while(s[i]==c && i>0)
		i--;

	if (i<0 || s[i]==c)
		return(pos);

	return (i);
}


/**********************************************************************
 *
 *  edits a text string at position x, y
 *
 * s - the string
 * x - column position
 * y - line position
 *
 **********/

void Menu::editalphastr(char *s, byte column, byte line) {

	int step=columns/3;
	byte pos =0;
	byte scroll=0;
	boolean quit=false;
	boolean redrawchar=false;
	boolean scrolled=true;
	byte origin, ndx, i;

	char t[MENUSTRLEN];

	displaylocked=true;
	strcpy(t,s);

	do {

		if(redrawchar) {

                // 1st check wether we are getting out of bounds

			if(columns<=column+pos-scroll) {	// shift left
				scroll+=step;
				scrolled=true;

			}

			if(column+pos-scroll<0) {			// shift right
				scroll=max(0,scroll-step);
				scrolled=true;
			}


                // then redraw character

			if(scrolled) {
				origin=column-scroll;
				ndx=0;
				if(origin<0){
					ndx=-origin;
					origin=0;
				}

				lcd.setCursor(origin,line);
				for(i=0; (i<column) && s[ndx+i];i++)
					lcd.write(s[ndx+i]);

			}
			lcd.setCursor(column+pos-scroll,line);
			lcd.write(s[pos]);
			redrawchar=false;
		}

		lcd.setCursor(column+pos-scroll,line);
		lcd.blink();

		switch(getkey()) {

			case UPKEY:
				s[pos]=inrange(s[pos]+1,0x21,'z');
				redrawchar=true;
				break;

			case DOWNKEY:
				s[pos]=inrange(s[pos]-1,0x21,'z');
				redrawchar=true;
				break;

			case LONGUPKEY:
				if((s[pos+1]))
					pos++;
				break;

			case LONGDOWNKEY:
				pos=max(pos-1,0);
				break;
                /*
                 case XTRALONGUPKEY:	// insert character
                 for(i=0;s[i];i++);
                 if(i<MENUSTRLEN) {
                 for(i=i+1;i>pos;i--)
                 s[i]=s[i-1];
                 s[pos]=' ';
                 scrolled=true;
                 redrawchar=true;
                 }
                 break;

                 case XTRALONGDOWNKEY:	// delete character
                 for(i=pos;s[i];i++)
                 s[i]=s[i+1];
                 scrolled=true;
                 redrawchar=true;
                 break;
                 */

			case ENTERKEY:
				quit=true;
				break;

			case LONGENTERKEY:
			case XTRALONGENTERKEY:
				strcpy(s,t);
				quit=true;
				break;
		}
	} while (!quit);
	lcd.noBlink();

	displaylocked=false;
}


/**********************************************************************
 *
 *  edits a numeric string with a special char (decimal) and allows editing
 *
 * s - string holding values
 * x - column position where it is to be displayed
 * y - line position where it is to be displayed
 * decimal - character to serve as a decimal char, which is ignored when editing
 *
 **********/

void Menu::editnumstr(char *s, byte x, byte y, char decimal) {

	byte i,j,pos;
	boolean quit=false;
	boolean redrawchar=false;
	char t[MENUSTRLEN];

	strcpy(t,s);

	displaylocked=true;

	lcd.setCursor(x,y);
	lcd.print(s);

	for(pos=0;!isdigit(s[pos]);pos++);

	lcd.blink();

	do {

		if(redrawchar) {
			lcd.setCursor(x+pos,y);
			lcd.write(s[pos]);
			redrawchar=false;
		}

		lcd.setCursor(x+pos,y);

		switch(getkey()) {
			case UPKEY:             // increase by 1
				s[pos]=inrange(s[pos]+1,'0','9');
				redrawchar=true;
				break;

			case DOWNKEY:           // decrease by 1
				s[pos]=inrange(s[pos]-1,'0','9');
				redrawchar=true;
				break;

			case LONGUPKEY:         // move to right
				s[pos]=inrange(s[pos],'0','9');
				lcd.setCursor(x+pos,y);
				lcd.write(s[pos]);
				pos=skipright(s, pos, decimal);
				break;

			case LONGDOWNKEY:		// move to left
				s[pos]=inrange(s[pos],'0','9');
				lcd.setCursor(x+pos,y);
				lcd.write(s[pos]);
				pos=skipleft(s, pos, decimal);
				break;

			case XTRALONGUPKEY:     // insert character, keep total length of string constant,
                                    // so move out characters to right

				for(i=0;s[i];i++);  // goto end, at this point s[i]==0;

                for(i=i-1;i>pos;i--)
                    s[i]=s[i-1];
                s[pos]='0';
                lcd.setCursor(x,y);
                lcd.print(s);
                redrawchar=true;

				break;

			case XTRALONGDOWNKEY:	// delete character, keep total length of string constant
                                    // insert space at last string position

				for(i=pos;s[i+1];i++)   // move chars to the right 1 unit to left as long as next character is not 0, so do not copy 0
					s[i]=s[i+1];
                s[i]=' ';               // insert space before end of string
				pos=skipleft(s, pos, decimal);
				lcd.setCursor(x,y);
				lcd.print(s);
				redrawchar=true;

				break;

			case ENTERKEY:
				quit=true;
				break;

			case LONGENTERKEY:		// abort, restore original string
			case XTRALONGENTERKEY:
				strcpy(s,t);
				quit=true;
				break;
		}
	} while (!quit);

	lcd.noBlink();
	displaylocked=false;
}


/**********************************************************************
 *
 * fills string with char
 *
 **********/

void Menu::fillstring(char *s, char c) {
	byte i;

	for(i=0;i<MENUSTRLEN;s[i++]=c);
	s[MENUSTRLEN-1]=0;
}


/**********************************************************************
 *
 * force redraw at next call of display
 *
 **********/

void Menu::forceredraw() {
	mposold=NOTHING;

}


/**********************************************************************
 *
 * creates a menu instance for Flash String Storage
 *
 **********/

Menu::Menu(menuitem *m, byte size, PGM_P *ms, unsigned int (*serv)(int *args), void (*g)(byte n, char *s) ){

	menuentry=m;
	menusize=size;
	menustrings=ms;

	service=serv;
	getdynamicmenutext=g;
	texttype=1;

	menureset();
}

/**********************************************************************
 *
 * creates a menu instance for EEPROM String Storage
 *
 **********/

Menu::Menu(menuitem *m, byte size, streeng *eep, unsigned int (*serv)(int *args), void (*g)(byte n, char *s) ){

	eeprom=eep;
	menuentry=m;
	menusize=size;

	service=serv;
	getdynamicmenutext=g;
	texttype=2;

	menureset();
}


/**********************************************************************
 *
 * Initializes the menu panel
 *
 * s - points to char string to be printed at top of screen
 *
 **********/

void Menu::setpanel(char *s) {

	strcpy(head,s);
	head[MENUSTRLEN-1]=0;
}

/**********************************************************************
 *
 * disassembles the level into its byte constituents
 *
 * item - the menu item
 * i  - the position of the byte in the level
 *		mapping:
 *		ABCD -> pos (0) = A, pos (1) = B, pos (2) = C, pos (3) = D
 *
 *	returns A, B, C, or D
 *
 **********/



byte Menu::pos(unsigned int level, byte i) {
	return (level >> ((MENULEVEL-i-1)*4)) & 0x000F;
}


/**********************************************************************
 *
 * Initializes menuentry
 *
 * lvl - os the integer value indicating the place in the hierarchy 0x1234 would be 1.2.3.4 level
 * ndx  - index of PGM memory string for entry / or position in EEPROM space
 *
 **********/

int Menu::entry(unsigned int lvl,  byte ndx) {
	byte i, j, lowest;
	byte m[MENULEVEL];

	if(maxentry>=menusize)
		return 1;

	menuentry[maxentry].level=lvl;

	menuentry[maxentry].txt=ndx;
	menuentry[maxentry].submenu=NOTHING;

	if(mroot==STARTPOS)
		mroot=maxentry;
	else {
		if(menuentry[maxentry].level<menuentry[mroot].level)
			mroot=maxentry;
	}

	maxentry++;

	mpos=mroot;      // menu structure changed, so go back to root

	for(i=0; i<maxentry; i++){
		for(j=0; j<MENULEVEL; j++) {

			m[j]=pos(menuentry[i].level,j);

			if (m[j]!=0)
				lowest=j;  			// determine next submenu level
		}

		if (lowest<MENULEVEL-1) {
			m[lowest+1]=1;    			// and create fictious menu level
			menuentry[i].submenu=findentry(m);	// find it
		}
		else
			menuentry[i].submenu=NOTHING;		// if not no submenu

    }
	return 0;
}



/**********************************************************************
 *
 * Changes text of menuentry
 *
 * lvl - os the integer value indicating the place in the hierarchy 0x1234 would be 1.2.3.4 level
 * ndx  - index of PGM memory string for entry
 *
 * returns true if replaced, false if entry not found
 *
 **********/

boolean Menu::changetxt(unsigned int lvl, byte ndx) {

	byte i;

	for(i=0; i<maxentry; i++) {
		if(menuentry[i].level==lvl) {
			menuentry[i].txt=ndx;
			return true ;
		}
	}
	return false;
}

/**********************************************************************
 *
 * clears one line of the display
 *
 * line - the line
 *
 **********/

void Menu::clearline(byte line)
{
	byte i;

	if((line>=0) && (line<lines)) {
		lcd.setCursor(0,line);
		for(i=0;i<columns;i++)
			lcd.write(' ');
	}
}


/**********************************************************************
 *
 * clears the screen and returns to top level
 *
 **********/

void Menu::home()
{
	mpos=mroot;
	forceredraw();
}


/**********************************************************************
 *
 * resets all entries of the menutree
 *
 *
 **********/

void Menu::menureset(void) {

	head[0]=0;
	maxentry=0;			// delete all menuentries
	mroot=STARTPOS;			// means that there is initially no root
	displaylocked=false;		// empty menu must be displayed
	servicerequest=NOCOMMAND;		// eliminate all previous external service requests
	forceredraw();			// and force a redraw next time it is to be displayed
}


/**********************************************************************
 *
 * Initializes the Menu class
 *
 * up - pin for up key
 * enter - pin for enter key
 * down - pin for down key
 *
 * all pins should be pulled down and keys should provide contact to +5V
 *
 **********/

void Menu::begin(byte up, byte enter, byte down, byte light) {

	uppin=up;
	enterpin=enter;
	downpin=down;
    backlightpin=light;

	pinMode(uppin,INPUT);
	pinMode(enterpin,INPUT);
	pinMode(downpin,INPUT);

	digitalWrite(uppin,HIGH);		// switch on internal pull ups
	digitalWrite(enterpin,HIGH);
	digitalWrite(downpin,HIGH);

    columns = lcd.columns();
	lines = lcd.lines();

	home();
}


/**********************************************************************
 *
 * switches on backlight
 *
 * sec - number of seconds
 *
 **********/

void Menu::switchonbacklight(byte sec) {
    unsigned long now;

    if(backlightpin==0)
        return;

    now=millis();
    backlightend=now+sec*1000;      
    if(backlightend<now)            // rollover will occur during backlight period
        backlightend=0xFFFFFFFF;    // switch it off a bit earlier to avoid rollover

    digitalWrite(backlightpin,HIGH);
}


void Menu::testtoswitchoffbacklight(void) {

    if(backlightpin==0) // no backlight function
        return;

    if(millis()>backlightend)     // this could be executed a number of times, without problem
        digitalWrite(backlightpin,LOW);
}

/**********************************************************************
 *
 * finds menuentry, which corresponds to position array
 *
 * m - points to array of bytes indicating the level of the menu entry
 *
 **********/

byte Menu::findentry(byte *m) {
	byte i;
	unsigned int sought=0;

	for(i=0;i<MENULEVEL;i++) {
		sought=(sought<<4) & 0xFFF0;
		sought|=m[i];
	}

	for(i=0; i<maxentry; i++) {
		if(sought==menuentry[i].level)
			return i;
	}

	return NOTHING;
}


/**********************************************************************
 *
 * seeks next menuentry depending on search direction, returns number of entry or NOTHING if nothing found
 *
 * entry - for which entry are we seeking a successor / precedecessor / ...
 * dir - which direction are we seeking (LEVELUP, LEVELDOWN, SAMELEVELUP, SAMELEVELDOWN)
 *
 **********/

byte Menu::seek(int entry, byte dir) {

	byte i;
	byte level[MENULEVEL];

	int result=NOTHING;
	int lowest=0;

	for(i=0;i<MENULEVEL;i++) {
		level[i]=pos(menuentry[entry].level,i);

		if(level[i]!=0)
			lowest=i;
	}

	switch (dir) {
		case LEVELUP:
			level[lowest]=0;
			result=findentry(level);
			break;

		case LEVELDOWN:
			result=menuentry[entry].submenu;
			break;

		case SAMELEVELUP:
			level[lowest]=max(level[lowest]-1,1);
			result=findentry(level);
			break;

		case SAMELEVELDOWN:
			level[lowest]=min(level[lowest]+1,MAXLEVEL-1);
			result=findentry(level);
			break;

		default:
			result=NOTHING;
			break;
    }

	return result;
}




/**********************************************************************
 *
 *  handles menu events and returns menu entry code to process command
 *
 *  returns the menulevel of a menu where the enter key is pressed
 *
 **********/

unsigned int Menu::event() {

	byte next, lastfound;
	unsigned int command=0;

	do {
		display(true, true);

		next=NOTHING;

		switch(getkey())
		{
			case UPKEY:
            next=seek(mpos, SAMELEVELUP);
            break;

			case LONGUPKEY:
            next=seek(mpos, LEVELUP);
            break;

			case DOWNKEY:
            next=seek(mpos, SAMELEVELDOWN);
            break;

			case ENTERKEY:
            if (menuentry[mpos].submenu==NOTHING)
                command=menuentry[mpos].level;
            else
                next=seek(mpos, LEVELDOWN);
            break;

			case	LONGENTERKEY:
			case	XTRALONGENTERKEY:
            home();
            break;

			default:
            break;
		}

		if(next!=NOTHING)
			mpos=next;

		if(servicerequest!=NOCOMMAND) {	// we have encountered an external service request
			if(command==0) {		// a keyboard command has priority over external service requests
				command=servicerequest;		// so if no keyboard command, take the servicerequest as
				servicerequest=NOCOMMAND;		// a valid command and delete servicerequest content
			}
		}

	} while (command==0);

	return(command);
}


/**********************************************************************
 *
 * copies a string residing in
 *	PGMSpace
 *	EEPROM
 *	or dynmically generated into the string
 *
 * index - position of string in relevant memory, or if high bit is set, dynamic
 * s - string to receive the text
 *
 **********/


void Menu::obtainstring(char *s, byte index) {

	*s=0;

	if(index & 0x80)		// check if high bit is set, seek for string with lower 7 bits
		getdynamicmenutext(index & 0x7F, s);
	else {
		switch(texttype) {
			case 1:
				strcpy_P(s, (PGM_P)pgm_read_word(&(menustrings[index])));
				break;

			case 2:
				eeprom->read(s,index);
				break;
		}
	}


}
/**********************************************************************
 *
 * displays one menu entry on screen
 *
 * pos - the menu entry, which is to be displayed
 * line - the line of the display, where it is to be displayed
 *
 **********/

void Menu::display_entry(byte pos, int line) {
	char buffer[MENUSTRLEN];

	if((line<0) || (line>lines-1))  // lines out of bounds of screen
		return;

	if(pos==NOTHING)              // no valid entry
		return;

	clearline(line);

	if(pos==mpos) {			// show where we are
		lcd.setCursor(0,line);
		lcd.write('>');
	}

	obtainstring(buffer, menuentry[pos].txt);

	lcd.setCursor(1,line);
	lcd.print(buffer);

	if(menuentry[pos].submenu!=NOTHING) {		// indicate submenu available
		lcd.setCursor(columns-1,line);
		lcd.write('>');
	}
}



/**********************************************************************
 *
 * Finds menu entry on same menu level preceding the current one to display
 *
 **********/

byte Menu::findprevious(int current, int todisp) {

	byte maxprevious;
	byte previous, lastfound;
	byte i=0;

	previous=lastfound=current;

	maxprevious=max(todisp,1)*2/5;						// determines the look of the thing.

	do {
		previous=seek(lastfound,SAMELEVELUP);
		if (previous!=NOTHING)								// same thing here, if seek provides NOTHING needed
			lastfound=previous;								// not needed if not
		i++;
	} while ((i<maxprevious) && (previous!=NOTHING));

	return lastfound;
}


/**********************************************************************
 *
 * displays the panel
 *
 **********/

void Menu::display_panel() {
	lcd.home();
	lcd.print(head);
}


/**********************************************************************
 *
 * displays the menu on the screen
 *
 * panel - is there a panel on the top to be displayed ?
 * body - should the menu body be displayed ?
 *
 **********/

void Menu::display(boolean panel, boolean body) {
	
	byte current;
	int currline=0;
	
	if(mpos!=mposold) {
		mposold=mpos;
		
		lcd.clear();
		
		if (panel) {
			currline=1;
			display_panel();
		}
		
		if(body) {
			current=mpos;
			
			if (lines-currline>2)
				current=findprevious(mpos, lines-currline);
			
			do {				// display until end of screen or end of current submenu reached
				display_entry(current, currline++);
				current=seek(current, SAMELEVELDOWN);
			} while ((current!=NOTHING) && (currline<lines));
		}
	}
}  


