/***************************
 *
 *
 *	Library to create an interactive LCD screen menu system
 *
 *
 *	R. Kohleick	29.12.2011
 *
 * licensed under GNU GPLv3
 *
 **/

#ifndef Menus_h
#define Menus_h

#include "Arduino.h"
#include <myLiquidCrystal.h>
#include <avr/pgmspace.h>
#include <ctype.h>
#include <mystuff.h>
#include <streeng.h>


#define NOTHING			0xff			// indicates that there is no submenu

#define LEVELUP				1
#define LEVELDOWN			2
#define SAMELEVELUP			3
#define SAMELEVELDOWN		4

#define LONGKEYOFFSET		4
#define XTRALONGKEYOFFSET	7

#define NOKEY				0

#define UPKEY				1
#define DOWNKEY				2
#define ENTERKEY			3
#define LONGUPKEY			5	// == UPKEY + LONGKEYOFFSET
#define LONGDOWNKEY			6	// == DOWNKEY + LONGKEYOFFSET
#define LONGENTERKEY		7	// == ENTERKEY + LONGKEYOFFSET
#define XTRALONGUPKEY		8	// == UPKEY + XTRALONGKEYOFFSET
#define XTRALONGDOWNKEY		9	// == DOWNKEY + XTRALONGKEYOFFSET
#define XTRALONGENTERKEY	10	// == ENTERKEY + XTRALONGKEYOFFSET
#define HOTKEYUD			11	// up and down together
#define HOTKEYUE			12	// up and enter together
#define HOTKEYDE			13	// down and enter together


#define MENUSTRLEN			21	// length of a menu string, needs to be 
								// width of display + 1

#define MENULEVEL			4	// we have maximum 4 nested levels
#define MAXLEVEL			16	// maximum of entries for each level

#define MAXENTRIES			43	// maximum number of entries for menutree

#define UPPIN				10
#define ENTERPIN			9
#define DOWNPIN				8	// pins for sensing keyboard


/********
 *
 * command constants for Menu event handling
 *
 **/

#define SYNTAXERROR		0xfffe
#define NOCOMMAND		0xffff			// not a valid command or reference to meny entry

#define ALERT_EVENT		0x00f0			// Alert attempts to send a message
#define LOG_EVENT		0x00f1			// Controler attempts to output logging message / status
#define SMS_EVENT		0x00f2			// sms received
#define CALL_EVENT		0x00f3			// telcall received


/***********
 *
 * and now the menu relevant structures
 *
 **/

int const menumaxarg=8;			// maximum number of arguments supplied to a menu command

struct menuitem {
	unsigned int level;			// indicates at which menu level we are, each nibble of the (2) bytes holds an entry
	byte txt;					// the text for the Menustring, indicates position in array of pgm memory strings or EEPROM strings,
								// if high bit set, seek dynamic string generation
	byte submenu;				// the array index number of a submenu, if any
};



class Menu {
	
public:
	
	Menu(menuitem *m, byte size, PGM_P *ms=0, unsigned int (*serv)(int *args)=0, void (*string)(byte n, char *s)=0  );
	
	// an array of menuitems, as well as an indicuation of the menusize
	//
	// plus an array of pointers to the (static) menustrings in programspace
	//
	// as well as a function, which is called while waiting for
	// keyboard input
	//
	// additionally, provide a function which copies dynamic menu entries
	// into a string, so that the program can modify the menu
	
	Menu(menuitem *m, byte size, streeng* eep=0, unsigned int (*serv)(int *args)=0, void (*string)(byte n, char *s)=0  );
	
	// an array of menuitems, as well as an indicuation of the menusize
	//
	// plus the menustrings in EEPROM
	//
	// as well as a function, which is called while waiting for
	// keyboard input
	//
	// additionally, provide a function which copies dynamic menu entries
	// into a string, so that the program can modify the menu
	
	void menureset(void);			// eliminate all menu entries, reset menu
	
	void begin(byte up, byte enter, byte down, byte backlight=0);		// start the Menu by indicating the keys used for navigating
	
	boolean displaylocked;			// if true, no subroutine may write to display,
									// so that we do not disturb the cursor;
	
	byte getkey(void);				// gets key
	boolean enterkey_pressed(void)	{return (digitalRead(enterpin)==LOW); };
	boolean upkey_pressed(void)		{return (digitalRead(uppin)==LOW); };
	boolean downkey_pressed(void)	{return (digitalRead(downpin)==LOW); };

    void switchonbacklight(byte secs);  // switches on the backlight for secs

	void home(void);						// go to menu root
	void clearline(byte l);					// clears line l of display
	void cleararound(byte l) { clearline(l-1); clearline(l); clearline(l+1); };	// clears around the lth line
	
	int entry(unsigned int lvl, byte i);			// enters menu entry
	boolean changetxt(unsigned int lvl, byte i);	// modifies entry text
	
	void forceredraw(void);					// forces redraw of display at next display() call
	void setpanel(char *s);					// sets the panel
	void display_panel(void);
	void display(boolean h, boolean b);		// displays panel & body
	
	unsigned int event(void);				// waits for keyboard event & handles menus
	
	void editnumstr(char *s, byte column, byte line, char spec);	// edit numerical string s at screenpos column, line with special char spec
	void editalphastr(char *s, byte column, byte line);				// edit alpha string s at position column line
	
	int getint(int initial, int low, int high, byte field, byte dec, char *q, byte x, byte y);
	int getenum(int initial, int low, int high, char *s[], byte x, byte y);
	
	boolean gettime(byte *h, byte *m, char *q, byte x, byte y);				// allows for an input of time
	boolean getdate(byte *d, byte *m, int *yr, char *q, byte x, byte y);	// allows for an input of date
	
	int args[menumaxarg];			// arguments passed from external serviceroutine, 
	
	void obtainstring(char *s, byte index);		// copies string stored in EEPROM, PRGMSPACE or dynmic to s

private:
	menuitem *menuentry;			// the entries - as arduino does not support new(), calling program allocates the size
	
	byte texttype;				// indicates, which method to store strings is used (1=PGM SPACE, 2=EEPROM
	
	PGM_P		*menustrings;	// strings for the menu reside in programspace
	streeng		*eeprom;		// strings may reside in EEPROM
	
	unsigned int (*service) (int *a);		// function to service during keyboard waits, should not be too long & to be provided by calling prgm
								// if not provided by calling program, no effect
								// the function returns a valid command entry code (similar to menu position), which is passed through via menu.event
								// and saves parameters in the parameter array, which may be then evaluated by calling program.
	
	void externalservice(void);			// handles the external calling function
	unsigned int servicerequest;		// which may ask for a certain request stored here
	
	void (*getdynamicmenutext) (byte nr, char *s);	// copies string for dynamic menu entry nr to string s
	
	byte pos(unsigned int lvl, byte i);	// disassembles level into position constituents
	
	byte uppin;						// the pins for up, down and enter keys and backlight
	byte enterpin;
	byte downpin;
	byte backlightpin;

    void testtoswitchoffbacklight();
    unsigned long backlightend;     // time when backlight should be switched off
    
	byte maxentry;					// maximum entry, keeps track of entries
	byte menusize;					// maximum size of the menu
	
	byte mpos;				// where are we in current menu tree ?
	byte mposold;			// menu position of previous screen, indicates whether the display been changed
	byte mroot;				// this is the root = lowest menu level
	
	char head[MENUSTRLEN];			// the top line of the display, contains status information
	
	byte columns, lines;			// dimension of display
	
	byte sensepin(byte pin, byte key);		// keyboard polling, any key pressed ?
	
	byte seek(int pos, byte dir);				// seeks next menu entry direction dir starting with pos
	byte findentry(byte *p);					// finds entry, which corresponds to position p
	byte findprevious(int current, int todisp);	// find preceding entry to display
	void display_entry(byte pos, int line);		// display one menu entry
	
	void fillstring(char *s, char c);			// fills string with char
	byte skipright(char *s, byte pos, char c);	// two routines to navigate inside a numeric string
	byte skipleft(char *s, byte pos, char c);
	
	int inrange(int c, int min , int max);		// forces char to range, permits a 'rollover', i.e. if <min -> max, if > max -> min
	
	void prepare_edit(byte *x, byte *y, char *q, char *s);	// prepares the editing of string s with question at pos x, y
	void normaltimechar(char c, char *s, byte pos);		// normalizes chars in time string
	boolean correctdate(byte day, byte mon, int year);	// determines if date is correct
	
};

#endif
