/*****************
 *
 *	Module to control the DS2415 One Wire Real Time Clock
 *	
 *	Reinhard Kohleick, May 3rd, 2012
 *	
 *	The DS 2415 only counts seconds, the current version attempts to 
 *	provide the same functionality as the DS1307 Clock. The module should
 *	work properly (day of week) during the 21st century.
 *
 *	library should work properly until 2136
 *	
 *	Spares you footprint on a PCB (DS2415 is smaller)
 *	and eliminates need for additional 2 Wire library, pins, and overhead
 *	in case you already have a one wire device running (e.g. DS18B20)
 *	
 *	The software is issued under the
 *
 *		GNU GENERAL PUBLIC LICENSE
 *		Version 3, 29 June 2007
 *
 * Files:
 *	RTC_DS2415.h
 *	RTC_DS2415.cpp
 * 
 */

#include <RTC_DS2415.h>
#include <ctype.h>

const byte secmins=60;
const unsigned int sechours=3600;
const unsigned long secdays=86400;

/*****************/


const unsigned int fouryeardaycount=1461; // 4*365+1

    //#define SWITCHYEAR(y) (!((y)&0x03) && ((y)!=100)) // 2100 is no switchyear, other than that %4=0
#define SWITCHYEAR(y) !((y)&0x03)
    
/***************
 *
 * determines days of all years preceding current rtc 
 *
 ***/


unsigned int Clock::daysofpreviousyears(byte year) {
	unsigned int days=0;
    byte fouryearperiods;

    fouryearperiods=year>>2;
	days=fouryeardaycount*fouryearperiods;
    
    year&=0x03;
    
	days+=365*year;
    if(year>0)      // we are in year 1 or later of the current switchyear period
        days++;

	return days;
}

/***************
 *
 * determines year from long int seconds
 *
 ***/

byte Clock::yearfromsec(unsigned long time) {
	unsigned int days;
	byte year;
	
    days=time/secdays;      // 0    .. 1460 -> 1st 4 year period && we are already in 2011
                            // 1461 .. 2921 -> 2nd
                            // 2922 .. 4382 -> 3rd

	year=days/fouryeardaycount;
    year*=4;

	days%=fouryeardaycount;
	
    // now we have year 0 = [0..365], 1 = [366+0 .. 366+364], 2 = [366+365 .. 366+365+364], 3 = [366+2*365 .. 366+2*365+364] days

    if(days<366)
        return year;
	if(days<366+1*365)
        return year+1;
 	if(days<366+2*365)
        return year+2;
	return year+3;
}

/***************
 *
 * calculates days of all months preceding current mon
 *  
 *  month = 0 .. 11
 *
 ***/

    // month   jan                                               dec
    // ndx     0   1   2   3    4    5    6    7    8    9   10   11   12

int month[]= { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365};

int Clock::daysofpreviousmonths(byte year, byte mon) {

    int days=month[mon];

	if(SWITCHYEAR(year) && mon>1)  // we are later than in feb. in a switchyear
		days++;

	return days;
}


/***************
 *
 * determines secs, min, hours, days etc. from long int
 *
 ***/

void Clock::convertsecstortc(unsigned long time) {
    
	unsigned int days, previousdays;
	byte i;
      
	rtc[DS2415_YEAR]=yearfromsec(time);

	previousdays=daysofpreviousyears(rtc[DS2415_YEAR]);  // calculate, how many days are there in years preceding this one

	rtc[DS2415_SEC]=time%60;    // obtain seconds
	time/=60;
	rtc[DS2415_MIN]=time%60;    // minutes
	time/=60;
	rtc[DS2415_HOUR]=time%24;   // and hours
    
	days=time/24;               // this is the number of days passed since start of year 0, where 00:00:00 on 1.1.2000 = 0
      
    days-=previousdays;         // above we calculated how many day are there in the previous years, subtract to obtain number of days in this year

                                // now we have to search for the right month for this day in the year
    for(i=days/32;days>=daysofpreviousmonths(rtc[DS2415_YEAR],i);i++);

    rtc[DS2415_MON]=i-1;        // i now is the next month, i.e. i is the first month, where days < days of previousmonths, so we are in the i-1 th month
    rtc[DS2415_DAY]=days-daysofpreviousmonths(rtc[DS2415_YEAR],rtc[DS2415_MON]);  // day of month is then obtained by subtracting days in previous months
    rtc[DS2415_DOW]=calcdow();      // and finally determine day of week
}



/***************
 *
 * determines long int from secs, min, hours, days etc
 *
 ***/

unsigned long Clock::convertrtctosec(byte *rtc) {
	unsigned long result;
	unsigned int days;
	
	days=daysofpreviousyears(rtc[DS2415_YEAR]);                     // all days in the years preceding current year
	days+=daysofpreviousmonths(rtc[DS2415_YEAR], rtc[DS2415_MON]);  // all days in this year in the months preceding current month
	days+=rtc[DS2415_DAY];                                          // all days in the current month up to and including yesterday (as it is calendar day-1)

	result=secdays*days;
	result+=rtc[DS2415_HOUR]*(long)sechours;                        // plus all seconds of today
	result+=rtc[DS2415_MIN]*(long)secmins;
	result+=rtc[DS2415_SEC];
	
	return result;
}


/***********************************
 *
 *  save one wire address of clockchip to private
 *
 */
void Clock::begin(byte* a) {	
	address=a;
}


/**************
 *
 *  sets the one wire clock
 *
 */

void Clock::setDs2415(void) {
	unsigned long time;
	
	time=convertrtctosec(rtc);
	writeds2415onewire(time);	
}


/**************
 *
 *  reads the one wire clock
 *
 */

void Clock::readDs2415(void) {
	
	seccount=readds2415onewire();
	convertsecstortc(seccount);
}  



/**************
 *
 *  writes rtc to DS2415 via onewire
 *
 **/

void Clock::writeds2415onewire(unsigned long secs) {
	
	onewire->reset();
	onewire->select(address);

	onewire->write(0x99,1);     // 2415 write command
	onewire->write(0xAC);       // rtc start clock oscillator command
	
    onewire->write(secs);  
	secs>>=8;
    onewire->write(secs);  
	secs>>=8;
    onewire->write(secs);  
	secs>>=8;
	onewire->write(secs);  
}


/**************
 *
 *  reads rtc from DS2415 via onewire
 *
 **/

unsigned long Clock::readds2415onewire(void) {
	byte i;
	unsigned long result;
	byte *p;
	
	p=(byte *) & result;
	
	onewire->reset();
	onewire->select(address);
	onewire->write(0x66,1);		// read RTC command
	onewire->read();            // status byte, but we do not care
	
	for ( i = 0; i < 4; i++) {
		p[i] = onewire->read();	// data bytes constituting unsigned long, byte order verified
	}
	return result;
}




/**************
 *
 *  determines dow from date in rtc
 *
 *  Devised by Tomohiko Sakamoto in 1993, it is accurate for any date in the range October 15, 1582–December 31, 9999:
 * 
 * returns: 0 = Sunday, 1 = Monday, ....
 *
 */

byte Clock::calcdow(void) {
    
	byte t[]={ 0, 3, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4};
	int y;

    y=baseyear+rtc[DS2415_YEAR];
    
    y-=(rtc[DS2415_MON]<2);
    
	return ( y + y/4 - y/100 + y/400 + t[rtc[DS2415_MON]] + rtc[DS2415_DAY]+1 ) % 7;
}



/**************
 *
 *  prints the dat to string s
 *
 *  s - the string
 *
 ***/

#define ascii(c) ('0'+(c))

void myitoa(byte i, char *s) {
    s[0]=ascii(i/10);
	s[1]=ascii(i%10);    
}

void Clock::sprintdate(char *s) {
    byte y;
	
    myitoa(rtc[DS2415_DAY]+1,s);
	s[2]='.';
    myitoa(rtc[DS2415_MON]+1,&s[3]);
	s[5]='.';
	s[6]='2';
    s[7]=ascii(rtc[DS2415_YEAR]/100);
    myitoa(rtc[DS2415_YEAR]%100,&s[8]);
	s[10]=(char)0;
}


/**************
 *
 *  prints the time to string s
 *
 *  s - the string
 *
 */

void Clock::sprinttime(char *s) {
	
    char *dayofweek[]={ "Su", "Mo", "Tu", "We", "Th", "Fr", "Sa"};
    byte dow;

    dow=rtc[DS2415_DOW];
	s[0]=dayofweek[dow][0];
	s[1]=dayofweek[dow][1];
	s[2]=' ';
    myitoa(rtc[DS2415_HOUR],&s[3]);
	s[5]=':';
    myitoa(rtc[DS2415_MIN],&s[6]);
	s[8]=(char)0;
}


/**************
 *
 *  Constructor of clock instance
 *
 **/

Clock::Clock(OneWire *ow) {
	
	onewire=ow;
}


/**************
 *
 *  set time to ds2415
 *
 * input    h: hours [0..23]
 *          m: minutes [0..59]
 *
 **/

void Clock::settime(byte h, byte m) {
		
	rtc[DS2415_SEC]=0;
	rtc[DS2415_MIN]=m;
	rtc[DS2415_HOUR]=h;
	
    setDs2415();
}



/**************
 *
 *  set date to ds2415, preserve time as closely as possible
 *
 * input    d: calendar day [1..31]
 *          m: calendar month [1..12]
 *          y: year [0..2100]
 *
 **/

void Clock::setdate(byte d, byte m, int y) {
	
	y-=baseyear;
	
	readDs2415();
	
	rtc[DS2415_DAY]=d-1;
	rtc[DS2415_MON]=m-1;
	rtc[DS2415_YEAR]=y;
	
	setDs2415();
}
