/*****************
 *
 *	Module to control the DS2415 One Wire Real Time Clock
 *	
 *	Reinhard Kohleick
 *	
 *	The DS 2415 only counts seconds, the current version attempts to 
 *	provide the same functionality as the DS1307 Clock. The module should
 *	work properly (day of week) during the 21st century.
 *	
 *	Spares you footprint on a PCB (DS2415 is smaller)
 *	and eliminates need for additional 2 Wire library, pins, and overhead
 *	in case you already have a one wire device running (e.g. DS18B20)
 *	
 *	The software is issued under the
 *
 *		GNU GENERAL PUBLIC LICENSE
 *		Version 3, 29 June 2007
 *
 * Files:
 *	RTC_DS2415.h
 *	RTC_DS2415.cpp
 *
 */

#ifndef RealtimeClock2415_h
#define RealtimeClock2415_h

#include "Arduino.h"
#include "OneWire.h"

#define DS2415_SEC	0
#define DS2415_MIN	1
#define DS2415_HOUR 2
#define DS2415_DOW	3
#define DS2415_DAY	4
#define DS2415_MON	5
#define DS2415_YEAR	6

const int baseyear=2000;    // start of time measurement, must be switchyear


class Clock {
	
public: 
	
	void updateclock(void)	{ readDs2415(); };
	
	void sprinttime(char *s);
	void sprintdate(char *s);

	byte            seconds(void)	{ return rtc[DS2415_SEC]; };
	byte            hours(void)     { return rtc[DS2415_HOUR]; };
	byte            minutes(void)	{ return rtc[DS2415_MIN]; };
	byte            dow(void)       { return rtc[DS2415_DOW]+1; };          // 1=Sunday, 2=Monday ... 7=Saturday
	byte            day(void)       { return rtc[DS2415_DAY]+1; };          // 1 ... 28, 29, 30, 31
	byte            mon(void)       { return rtc[DS2415_MON]+1; };          // 1 ... 12
	int             year(void)      { return rtc[DS2415_YEAR]+baseyear; };  // 2000 ... 2100
    unsigned long   now(void)       { return seccount; };
	
	void settime(byte h, byte m);
	void setdate(byte d, byte m, int y);

	void begin(byte *p);
	
	Clock(OneWire *ow);

private:
	byte *address;	// address of one 2415 wire device

    unsigned long seccount; // holds raw data: counted seconds from clock
    
    OneWire *onewire;

	void setDs2415(void);
	void readDs2415(void);

    byte rtc[7];	// holds time data
    void convertsecstortc(unsigned long time);
	unsigned long convertrtctosec(byte *p);

	byte yearfromsec(unsigned long time);
	
	unsigned int daysofpreviousyears(byte y);
	int daysofpreviousmonths(byte y, byte m);

	void writeds2415onewire(unsigned long secs);
	unsigned long readds2415onewire(void);
	
	byte calcdow(void);
};

#endif

