/***************************
 *
 *
 *	Library to access GSM Modems
 *
 *  tested for
 *  
 *  Siemens tc35i       19200 baud
 *  Siemens mc35i       19200 baud
 *  Sagem   HiLo       115200 baud
 *
 *
 *	R. Kohleick	29.12.2011
 *
 * licensed under GNU GPLv3
 *
 **/

#include <ctype.h>
#include <mystuff.h>
#include <gsmmodem.h>
#include "Arduino.h"

char okstring[]={ 'O', 'K', 0x0D, 0x0A, 0x00};


/*****************************
 *
 * initializes gsm variable
 *
 *
 *****/

GSMModule::GSMModule(char *buf, byte size, char *tel) {
	
	buffer=buf;
	buffsize=size;
	telno=tel;
    available=false;
}

/*****************************
 *
 *	checks, whether we have incoming call or message
 *
 *	returns :   0 - no call or message
 *             -1 - incoming call
 *             n>0 - incoming sms at location n
 *
 *****/

int GSMModule::incoming(void) {
    
    if(Serial.available()==0)
        return 0;
    
    mydelaylong(2*mediumdelay);             // give the line some time to receive the message

    if(readln(buffer,buffsize)<0)           // and read it
        return 0;
    
    truncleft(buffer);
    truncright(buffer);
    
    if(strcmp(buffer,"RING")==0)
        return -1;

    parseline(buffer);
    
    if(argc==3 && strcmp(argv[0],"+CMTI")==0) {    // we are only looking for unsolicited CMTI messages with valid 3 arguments here
        return atoi(argv[2]);                      // return SMS index
    }
    
    return 0;   // space to accomodate further CIEV modes
    }

/*****************************
 *
 * copies network operator into buffer
 *
 *****/

void GSMModule::networkoperator(void) {
    
	sendcommand("AT+COPS?");	// check for operator name
	strcpy(buffer,argv[3]);		// and copy it to buffer
}


/*****************************
 *
 *	checks, whether we have network service
 *
 *	returns true: registered with network
 *          false: not registered
 *
 *****/

boolean GSMModule::networkregistered(void) {
    
	sendcommand("AT+CREG?");	// check for operator name
    return (*argv[2]=='1' || * argv[2]=='5'); // 1 - home network, 5 - roaming
}


/*****************************
 *
 *	starts the gsm module
 *
 *	may take a long time
 *	returns
 *		0	all ok
 *		2	not correct mobile version
 *		3	syntax error in message from mobile
 *		4	no network service
 *      5   no space in mobile module to accept SMS
 *  
 *  indicates whether mode is active
 *
 *****/

const byte maxnetworkwait=20;

byte GSMModule::startmodule(void) {
	byte cnt=0;
    
    Serial.flush();
    
	sendcommand("ATE0");            // switch off local echo
    
    /*
     if(sendcommand("ATI")<0)       // check for right version
     return 3;
     
     if( strcmp(argv[1],"TC35i")!=0 && strcmp(argv[1],"MC35i")!=0 )
     return 2;
     */
    
	sendcommand("AT+CMGF=1");       // switch to sms text mode   
    sendcommand("AT+CPMS=ME");      // set preferred message storage to mobile
                                    // this works for the Sienems modules TC35i and MC35i for others,
                                    // this could be alternatively be
                                    //  sendcommand("AT+CPMS=\"ME\",\"ME\",\"ME\"")
	maxsmsstorage=atoi(argv[2]);    // and store maximum number of possible messages
    if(maxsmsstorage==0)            // we need sms storage, so return with error if there is no space
        return 5;
    sendcommand("AT+CNMI=1,1");
 	deleteallsms();					// delete all sms in memory, if any

    do {                            // check whether we are registered with a network ?
        delay(largedelay);
    } while(cnt++<maxnetworkwait && !networkregistered());

    if(cnt==maxnetworkwait)         // no, we are not - return error
        return 4;

	networkoperator();              // yes, get the oprator name
	return 0;
}

/*****************************
 *
 *	starts the gsm module
 *
 *	may take a long time
 *	returns
 *		0	all ok
 *		1	no mobile connected / timeout
 *		2	not correct mobile version
 *		3	syntax error in message from mobile
 *		4	no network service
 *      5   no sms storage space
 *
 *****/

byte GSMModule::begin(void) {
	int res;
    
	Serial.flush();
    
	mydelaylong(10000);      // give modem enough time to power up	
    
    Serial.setTimeout(timeout);
    
    available=false;
	if(sendcommand("AT")<0)   
		return 1;                // timeout error / communication problem with modem, modem not available

    res=startmodule();
    if(res==0)
        available=true;
    
	return res;
}

/*****************************
 *
 * waits for occurence of char in serial
 *
 * returns:
 *
 *	1 if timeout occured, i.e. no character available from Serial port
 *	0 if found
 *
 ****/

void GSMModule::waitforchar(char pattern) {
	int c;
	
	do {
		c=Serial.read();
	} while ((c>=0) && (pattern!=c));
}

/*****************************
 *
 * reads line terminated by CR into buffer
 *
 * returns length of read string
 * or negative value in case of error
 *
 ****/

int GSMModule::readln(char *b, byte bsize) {
    byte c;
    char *p;
    
	c=Serial.readBytesUntil(0x0a, b, bsize);        // line is terminated by 0x0d, 0x0a pairs

    b[c]=0;

    if(c==0)            // no characters read or timeout
        return -1;
    
    truncright(b);      // line read, but this contains the 0x0d character, delete end of line 0x0d
    return strlen(b);   // return length of line
}


/*****************************
 *
 * parses line of parameters passed by module
 *
 * returns
 *	n>0 number of parameters passed, also stored in argc
 *	in buffer: n zero terminated strings, each pointed at by 
 *	argv array element
 *
 ****/

byte GSMModule::parseline(char *b) {
	byte i;
	argc=0;
	
	while(*b && (argc<maxargs)) {
		
		while(isspace(*b))		// skip leading spaces
			b++;
		if (*b) {               // if not end of line, we have a message to process
			if(argc==0 && (*b=='+' || *b=='^')) {	// 1st field in message, check if command field
				argv[argc++]=b;				// remember begining of command field
				while(*b!=':' && *b)		// terminating character for first field is ':' or EOLN, go there and beyond to beginning of next field
					b++;
			} else {                        // not the first field, rather arguments to the command field
				if(*b=='"') {				// check if text field
					b++;					// go one further for start of text
					argv[argc++]=b;			// remember this
					while(*b!='"' && *b)	// then, go to end of text field
						b++;
					if(*b=='"')
						*b++=0;             // and one beyond, but change that char to eoln, stripping the text field of its "
					
				} else      // not a text field, just remember beginning
					argv[argc++]=b;
                
                while(*b!=',' && *b!=0x0d && *b)	// seek terminating character for every field other than the first
                    b++;
			}
            if(*b)         // if this is not end of line, terminate this substring and go one beyond
                *b++=0;
		}
	}
	for(i=0;i<argc;i++) {       // strip strings identified of leading and traling white spaces
        truncleft(argv[i]);
        truncright(argv[i]);
	}	
    return argc;                // and return the number of arguments identified
}

/*****************************
 *
 * sends command to GSM Module
 *
 * returns
 *	0 if response just 'OK'
 *	n if there are some results returned, parameters in buffer as 0 terminated strings, referred to by argv, argc
 *	-1 timeout after command execution
 *  -2 line could not be read
 *	-3 ERROR returned
 *	-4 answer was to long to fit in buffer
 *
 *****/

const int maxcommandwait=100;

int GSMModule::sendcommand(char *cmd) {
	int i=0;
	int size;
	int timeout;
	char *p, *last;
	
    Serial.flush();        // eliminate URC issued before command
	Serial.println(cmd);   // issue command
	
	last=p=buffer;
	size=buffsize;
    *last=0;
	
    do {
        mydelaylong(smalldelay);
    } while(Serial.available()==0 && i++<maxcommandwait);    // wait for some response
    
	if(i==maxcommandwait) {     // no response at all within waiting period, return a time out error
        i=-1;
        *buffer=0;
        goto terminate;
    }
    
 	waitforchar(0x0a);          // we have a response, now wait for 1st LF
    
	while(size>0) {             // and begin to receive line while buffer has enough space
		if(readln(p,size)<0) {  // timeout error, do not wait for new chars as the line is down
			*buffer=0;
            i=-2;
            goto terminate;
		}
        
		if(strcmp(p,"ERROR")==0) {	// ERROR encountered, return -3
			*buffer=0;
            i=-3;
            goto terminate;
		}
		
		if(strcmp(p,"OK")==0) {		// if ok encountered, ignore ok,
			*last=0;
			i=parseline(buffer);	// delay here assures that there are no timing issues with TC upon successful exit            
            goto terminate;         // now we have encountered a valid line, wait a bit and then process
		}
		
		if (*p) {					// no end of message received, accept (partial) line received to buffer
			for(i=0; *p ; p++)      // go to end of line and 
				i++;
			size=size-i-1;			// check whether we can store the length of line plus one CR
			if(size>=0) {
				last=p;				// yes, we can, remember current position as last end of line
				*p++=0x0d;			// insert CR and move ptr one further to receive next line
			}
		}
	}
	*last=0;
	
	Serial.find(okstring);          // could not read entire response, wait for hte next "OK"
    i=-4;
    
terminate:
    
    mydelaylong(postcommanddelay);
	return i;
}


/*****************************
 *
 * SMS command plus text to Module
 * number and text are 0 terminated strings
 *
 * returns waitfor OK result
 *
 *****/

const int smslength=160;
const int tempbuffsize=40;

byte GSMModule::sendsms(char *number, char *text) {
	byte res;
    
    text[min(smslength,strlen(text))]=0;            // truncate message at smslength, if longer
    
    Serial.flush();
	Serial.print("AT+CMGS=\"");
	Serial.print(number);
	Serial.println("\"");
	
	while(Serial.available()==0);
 	waitforchar(0x0a);          // we have a response, now wait for 1st LF

	Serial.find("> ");
	
	Serial.print(text);
	Serial.write(0x1A);
    
	while(Serial.available()==0);   // wait for first character to appear
 	waitforchar(0x0a);              // we have a response, now wait for 1st LF
 	waitforchar(0x0a);              // we have a response, now wait for 1st LF

	mydelaylong(mediumdelay);
	res=Serial.find(okstring);
	
    mydelaylong(postcommanddelay);

    if(res)
        return 0;
    return -1;
}

/*****************************
 *
 * deletes sms at storage index ndx (from preferred storage)
 *
 *****/

void GSMModule::deletesms(byte ndx) {
	char cmd[11];
	byte i=8;
    
	strcpy(cmd,"AT+CMGD=");
    if(ndx>9)
        cmd[i++]='0'+ndx/10;
    cmd[i++]='0'+ndx%10;
    cmd[i]=0;
	sendcommand(cmd);
    
    mydelaylong(mediumdelay);   // command needs time to reorder sms inside modem
}



/*****************************
 *
 * deletes all sms in preferred storage (Mobile)
 * 
 * returns
 *		0 if ok
 *		negative if error occured
 *
 *****/

void GSMModule::deleteallsms(void) {
	byte i;
    
	for(i=1;i<=maxsmsstorage && smswaiting()>0; i++) {
		deletesms(i);
    }
}

/*****************************
 *
 * reads sms at storage index ndx (from preferred storage)
 * 
 * returns
 *		n>0 number of sms in storage
 *       0 if no sms in storage
 *		<0 sendcommand result if error occured
 *
 *****/

int GSMModule::readsms(byte ndx) {
	char cmd[11];
    byte i=8;
    int res;
    
	strcpy(cmd,"AT+CMGR=");
    
    if(ndx>9)
        cmd[i++]='0'+ndx/10;
    cmd[i++]='0'+ndx%10;
    cmd[i]=0;
    res=sendcommand(cmd);
    
	if(res<=0) {         // some error or no valid sms
        *buffer=*telno=0;
        return res;
    }
    
    strcpy(telno, argv[2]);
    strcpy(buffer, argv[5]);
    return(ndx);
}

/*****************************
 *
 * reads first available sms from ME to buffers
 * 
 * returns
 *		index of first sms message in storage
 *		0 if error occured or nothing found
 *
 ****/

byte GSMModule::readfirstsms(void) {
    
    if(sendcommand("AT+CMGL=\"ALL\"")<0) {   // nothing found or error
//    if(sendcommand("AT+CMGL=ALL")<0) {   // nothing found or error
        *telno=*buffer=0;
        return 0;
    }
    
    mydelaylong(mediumdelay);                 // give modem some time to store message
    
    strcpy(telno, argv[3]);
    strcpy(buffer, argv[6]);
    return atoi(argv[1]);        
}

/*****************************
 *
 * how many sms in preferred storage (Mobile)
 * 
 * returns
 *		0 if no smses
 *		n number of smses
 *
 *****/

byte GSMModule::smswaiting(void) {

    if(sendcommand("AT+CPMS=\"ME\"")<0)             // some error
//  if(sendcommand("AT+CPMS=ME")<0)             // some error
        return 0;
    
    *buffer=0;
    return atoi(argv[1]);
}



/*****************************
 *
 * checks for incoming telcall
 * 
 * return	true if call is waiting
 *			false if no call
 *
 * calling telno copied to answer
 *
 *****/

boolean GSMModule::callwaiting(void) {
	int result;
	
	result=sendcommand("AT+CLCC");
    
	if(result<=0 || *argv[3]!='4')
		return false;
    
	strcpy(telno,argv[6]);
    
	mydelaylong(5000);			// wait some time to allow caller to hear rings ..
	sendcommand("ATH");
	return true;
}